;; Generated package description from phps-mode.el  -*- no-byte-compile: t -*-
(define-package "phps-mode" "0.4.51.0.20250428.55352" "Major mode for PHP with code intelligence" '((emacs "26")) :commit "ff5fbd617d8e67f14f8928fd21f4de56bead987a" :authors '(("Christian Johansson" . "christian@cvj.se")) :maintainer '("Christian Johansson" . "christian@cvj.se") :keywords '("tools" "convenience") :url "https://github.com/cjohansson/emacs-phps-mode")
